@@ -16,11 +16,12 @@ import sys
 
 # Third-party imports
 from ipykernel.ipkernel import IPythonKernel
+
+# Local imports
 from spyder_kernels.comms.frontendcomm import FrontendComm
+from spyder_kernels.py3compat import isidentifier, PY2
 
 
-PY2 = sys.version[0] == '2'
-
 # Excluded variables from the Variable Explorer (i.e. they are not
 # shown at all there)
 EXCLUDED_NAMES = ['In', 'Out', 'exit', 'get_ipython', 'quit']
@@ -55,7 +56,8 @@ class SpyderKernel(IPythonKernel):
             'get_namespace_view': self.get_namespace_view,
             'set_namespace_view_settings': self.set_namespace_view_settings,
             'get_var_properties': self.get_var_properties,
-            'set_sympy_forecolor': self.set_sympy_forecolor
+            'set_sympy_forecolor': self.set_sympy_forecolor,
+            'set_pdb_echo_code': self.set_pdb_echo_code,
             }
         for call_id in handlers:
             self.frontend_comm.register_call_handler(
@@ -65,6 +67,7 @@ class SpyderKernel(IPythonKernel):
 
         self._pdb_obj = None
         self._pdb_step = None
+        self._pdb_print_code = True
         self._do_publish_pdb_state = True
         self._mpl_backend_error = None
 
@@ -103,6 +106,14 @@ class SpyderKernel(IPythonKernel):
         if self._pdb_obj:
             self._pdb_obj.set_spyder_breakpoints(breakpoints)
 
+    def set_pdb_echo_code(self, state):
+        """Set if pdb should echo the code.
+
+        This might change for each pdb statment and is therefore not included
+        in pdb settings.
+        """
+        self._pdb_print_code = state
+
     # -- Public API ---------------------------------------------------
     # --- For the Variable Explorer
     def set_namespace_view_settings(self, settings):
@@ -224,6 +235,64 @@ class SpyderKernel(IPythonKernel):
         return iofunctions.save(data, filename)
 
     # --- For Pdb
+    def is_debugging(self):
+        """
+        Check if we are currently debugging.
+        """
+        return bool(self._pdb_frame)
+
+    def do_complete(self, code, cursor_pos):
+        """
+        Call PdB complete if we are debugging.
+
+        Public method of ipykernel overwritten for debugging.
+        """
+        if cursor_pos is None:
+            cursor_pos = len(code)
+        if self.is_debugging():
+            # Get text to complete
+            text = code[:cursor_pos].split(' ')[-1]
+            # Choose pdb function to complete, based on cmd.py
+            origline = code
+            line = origline.lstrip()
+            if not line:
+                return
+            stripped = len(origline) - len(line)
+            begidx = cursor_pos - len(text) - stripped
+            endidx = cursor_pos - stripped
+            if begidx > 0:
+                cmd, args, _ = self._pdb_obj.parseline(line)
+                if cmd == '':
+                    compfunc = self._pdb_obj.completedefault
+                else:
+                    try:
+                        compfunc = getattr(self._pdb_obj, 'complete_' + cmd)
+                    except AttributeError:
+                        compfunc = self._pdb_obj.completedefault
+            elif line[0] != '!':
+                compfunc = self._pdb_obj.completenames
+            else:
+                compfunc = self._pdb_obj.completedefault
+
+            def is_name_or_composed(text):
+                if not text or text[0] == '.':
+                    return False
+                # We want to keep value.subvalue
+                return isidentifier(text.replace('.', ''))
+
+            while text and not is_name_or_composed(text):
+                text = text[1:]
+                begidx += 1
+
+            matches = compfunc(text, line, begidx, endidx)
+
+            return {'matches': matches,
+                    'cursor_end': cursor_pos,
+                    'cursor_start': cursor_pos - len(text),
+                    'metadata': {},
+                    'status': 'ok'}
+        return super(SpyderKernel, self).do_complete(code, cursor_pos)
+
     def publish_pdb_state(self):
         """
         Publish Variable Explorer state and Pdb step through
